from django.dispatch import Signal

# These are used internally to hook up to the request finished signal.
finished = Signal(providing_args=['request'])
process = Signal(providing_args=['type', 'message', 'values', 'request'])

# This is the signal that you hook the warnings in the code up to.
warning = Signal(providing_args=['type', 'message', 'values'])



from django.utils.translation import ugettext as _

EXTRA_FIELDS = 'extra-form-fields'  # OWASP: RE6
MISSING_FIELDS = 'missing-form-fields'  # OWASP: RE5
WRONG_METHOD = 'wrong-method'  # OWASP: RE1, RE2, RE3 and RE4
SESSION_CHANGED = 'session-changed'  # OWASP: SE5 and SE6
UNEXPECTED_CHARACTER = 'unexpected-character'  # OWASP: RE8
WRONG_METHOD = 'wrong-method'  # OWASP: RE1, RE2, RE3 and RE4

trans = {
    EXTRA_FIELDS: _('Attempt to process form with extra values'),
    MISSING_FIELDS: _('Attempt to process form with missing values'),
    SESSION_CHANGED: _('Session data changed'),
    UNEXPECTED_CHARACTER: _('Unexpected character'),
    WRONG_METHOD: _('Wrong HTTP method'),
}



from functools import wraps

from django.http import HttpResponseNotAllowed
from django.utils.decorators import available_attrs
from django.utils.log import getLogger

from flags import WRONG_METHOD
from signals import warning


logger = getLogger('django.request')


def require_http_methods(request_method_list):
    """Like the Django decorators, but they also raise a warning."""
    def decorator(func):
        @wraps(func, assigned=available_attrs(func))
        def inner(request, *args, **kwargs):
            if request.method not in request_method_list:
                # Raise our warning.
                warning.send(sender=require_http_methods, flag=WRONG_METHOD,
                             message=u'%s not allowed' % request.method,
                             values=[request_method_list])
                logger.warning('Method Not Allowed (%s): %s',
                               request.method, request.path,
                               extra={
                                    'status_code': 405,
                                    'request': request
                               })
                return HttpResponseNotAllowed(request_method_list)
            return func(request, *args, **kwargs)
        return inner
    return decorator


require_GET = require_http_methods(['GET'])
require_GET.__doc__ = ('Paranoid decorator to require that a '
                       'view only accept the GET method.')

require_POST = require_http_methods(['POST'])
require_POST.__doc__ = ('Paranoid decorator to require that a '
                        'view only accept the POST method.')

require_safe = require_http_methods(['GET', 'HEAD'])
require_safe.__doc__ = ('Paranoid decorator to require that a '
                        'view only accept safe methods: GET and HEAD.')



import functools

from django.conf import settings

from cef import log_cef


def report(signal, message=None, flag=None, sender=None, values=None,
           request=None, **kwargs):
    g = functools.partial(getattr, settings)
    severity = g('CEF_DEFAULT_SEVERITY', 5)
    cef_kw = {'msg': message, 'signature': request.get_full_path(),
            'config': {
                'cef.product': g('CEF_PRODUCT', 'paranoia'),
                'cef.vendor': g('CEF_VENDOR', 'Mozilla'),
                'cef.version': g('CEF_VERSION', '0'),
                'cef.device_version': g('CEF_DEVICE_VERSION', '0'),
                'cef.file': g('CEF_FILE', 'syslog'),
            }
        }
    log_cef(message, severity, request.META.copy(), **cef_kw)



import re

from django.forms import Form, ModelForm

from flags import EXTRA_FIELDS, MISSING_FIELDS, UNEXPECTED_CHARACTER, trans
from signals import warning

# Spot chars below 32, but allow, \t (9), \r (13) and \n (10).
chars = range(0, 9) + range(11, 13) + range(14, 32)
low_chars = re.compile('|'.join(map(chr, chars)))


class Paranoid(object):

    def __init__(self, data=None, files=None, **kwargs):
        super(Paranoid, self).__init__(data=data, files=files, **kwargs)
        # We need to check for extra data when the form is created.
        data = data or {}
        extra = [k for k in data if k not in self.fields]
        if extra:
            self.warn(EXTRA_FIELDS, extra)

        for k, v in data.items():
            # This assumes all binary data is going through FILES which
            # are not covered by django-paranoia. There really shouldn't
            # be any binary data here at all right?
            self.detect_low(k)
            self.detect_low(v)

    def detect_low(self, data):
        if not isinstance(data, basestring):
            return

        if low_chars.search(data):
            warning.send(sender=self.__class__,
                         flag=UNEXPECTED_CHARACTER,
                         message='Unexpected characters')
            # Not sure if we should send through the data in the message.
            # Would that be bad?

    def warn(self, flag, data):
        klass = self.__class__
        msg = (u'%s: %s in %s' % (trans[flag], data, klass.__name__))
        warning.send(sender=klass, flag=flag, message=msg, values=data)

    def is_valid(self):
        # We can't tell what is missing until the end.
        result = super(Paranoid, self).is_valid()
        missing = []
        if not result:
            for k, errors in self.errors.items():
                field = self.fields.get(k)
                for error in errors:
                    if field and error == field.error_messages['required']:
                        missing.append(k)
                    else:
                        missing.append(k)

        if missing:
            self.warn(MISSING_FIELDS, missing)

        return result


class ParanoidForm(Paranoid, Form):
    pass


class ParanoidModelForm(Paranoid, ModelForm):
    pass



import logging
from threading import local

from django.conf import settings
from django.utils.importlib import import_module

from signals import finished, process, warning

log = logging.getLogger('paranoia')

_locals = local()


def setup():
    if not hasattr(_locals, 'signals'):
        _locals.signals = []
        return True


def add_signal(signal, **kw):
    setup()
    # Let's not pickle the sender.
    kw['sender'] = kw['sender'].__name__
    _locals.signals.append(kw)


def reset(**kw):
    setup()
    _locals.signals = []


def process_signals(signal, **kw):
    # We need to batch up all signals here and then send them when the
    # request is finished. This allows us to pass through the request
    # to the reporters, allowing more detailed logs.
    if setup():
        return

    for data in _locals.signals:
        process.send(request=kw['request'], **data)


def config(*args, **kw):
    try:
        reporters = getattr(settings, 'DJANGO_PARANOIA_REPORTERS', [])
    except ImportError:
        return

    for reporter in reporters:
        try:
            to = import_module(reporter).report
        except ImportError:
            log.error('Failed to register the reporter: %s' % reporter)
            continue

        # Each reporter gets connected to the process signal.
        process.connect(to, dispatch_uid='paranoia.reporter.%s' % reporter)

    # The warning signal is sent by forms, sessions etc when they
    # encounter something.
    warning.connect(add_signal, dispatch_uid='paranoia.warning')
    # The finished signal is sent by the middleware when the request is
    # finished. This then processes all the warning signals built up so far
    # on that request.
    finished.connect(process_signals, dispatch_uid='paranoia.finished')
